#include "xospipsv.h"

XOspiPsv_Config XOspiPsv_ConfigTable[] __attribute__ ((section (".drvcfg_sec"))) = {

	{
		"amd,versal2-ospi", /* compatible */
		0xf1010000, /* reg */
		0xbebba31, /* xlnx,ospi-clk-freq-hz */
		0x0, /* xlnx,is-cache-coherent */
		0x0, /* xlnx,ospi-mode */
		0x40d8, /* interrupts */
		0xe2000000, /* interrupt-parent */
		0x8 /* xlnx,ospi-bus-width */
	},
	 {
		 NULL
	}
};